@@ -8,22 +8,27 @@ namespace torch::inductor {
 
 // NOTICE: Following APIs are subject to change due to active development
 // We provide NO BC guarantee for these APIs
-class TORCH_API AOTIModelContainerRunnerCuda : public AOTIModelContainerRunner {
+// NOLINTNEXTLINE(cppcoreguidelines-special-member-functions)
+class TORCH_HIP_CPP_API AOTIModelContainerRunnerCuda
+    : public AOTIModelContainerRunner {
  public:
   // @param device_str: cuda device string, e.g. "cuda", "cuda:0"
   AOTIModelContainerRunnerCuda(
       const std::string& model_so_path,
       size_t num_models = 1,
       const std::string& device_str = "cuda",
-      const std::string& cubin_dir = "");
+      const std::string& cubin_dir = "",
+      const bool run_single_threaded = false);
 
-  ~AOTIModelContainerRunnerCuda();
+  ~AOTIModelContainerRunnerCuda() override;
 
-  std::vector<at::Tensor> run(std::vector<at::Tensor>& inputs);
+  std::vector<at::Tensor> run_impl(
+      std::vector<AtenTensorHandle>& input_handles,
+      void* stream_handle) override;
 
   std::vector<at::Tensor> run_with_cuda_stream(
-      std::vector<at::Tensor>& inputs,
-      at::hip::HIPStreamMasqueradingAsCUDA cuda_stream);
+      const std::vector<at::Tensor>& inputs,
+      const at::hip::HIPStreamMasqueradingAsCUDA& cuda_stream);
 };
 
 } // namespace torch::inductor
